/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file util/tcpio.c
 * @brief code for synchronized access to TCP streams
 * @author Christian Grothoff
 *
 * Generic TCP code for reliable, mostly blocking, record-oriented TCP
 * connections. GNUnet uses the "tcpio" code for trusted client-server
 * (e.g. gnunet-gtk to gnunetd via loopback) communications.  Note
 * that an unblocking write is also provided since if both client and
 * server use blocking IO, both may block on a write and cause a
 * mutual inter-process deadlock.
 *
 * Since we do not want other peers (!) to be able to block a peer by
 * not reading from the TCP stream, the peer-to-peer TCP transport
 * uses unreliable, buffered, non-blocking, record-oriented TCP code
 * with a select call to reduce the number of threads which is
 * provided in transports/tcp.c.
 **/

#include "config.h"

#ifndef CYGWIN
 #include <sys/socket.h>
#else
 #include <cygwin/socket.h>
 #include <fcntl.h>
#endif
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <netdb.h>
#include <unistd.h>

#include "util/logging.h"
#include "util/returnvalues.h"
#include "util/tcpio.h"
#include "util/xmalloc.h"

#define DEBUG_TCPIO 0

/**
 * Catch sigpipe on send, and do nothing
 **/
static void catcher(int sig) {
  LOG(LOG_MESSAGE,
      "MESSAGE: signal %d caught (ignored)\n", sig);
  signal(sig, catcher);
}

/**
 * Initialize TCPIO module.
 * Installs a signal handler for SIGPIPE because read/write
 * operations on sockets can send that signal and we would
 * like to just ignore it.
 **/
void initTCPIO() {
  if (SIG_ERR == signal(SIGPIPE, SIG_IGN)) {
    /* retry: install 'method'  as signal handler */
    if (SIG_ERR == signal(SIGPIPE, catcher))
      LOG(LOG_WARNING,
	  "WARNING: could not install handler for SIGPIPE!\n"\
	  "Attempting to continue anyway.");
  }
}

/**
 * Shutdown TCPIO module.
 **/
void doneTCPIO() {
  /* should we restore the signal handler here? */
}

/**
 * Initialize a GNUnet client socket.
 * @param port the portnumber in host byte order
 * @param ip IP of the host to connect to, in network byte order
 * @param result the SOCKET (filled in)
 * @return OK if successful, SYSERR on failure
 **/
int initGNUnetClientSocketIP(unsigned short port,
			     IPaddr ip,
			     GNUNET_TCP_SOCKET * result) {
  result->ip = ip;
  result->port = port;
  result->socket = -1; /* closed */
  result->outBufLen = 0;
  result->outBufPending = NULL;
  MUTEX_CREATE(&result->readlock);
  MUTEX_CREATE(&result->writelock);
  return OK;
}

/**
 * Initialize a GNUnet client socket.
 * @param port the portnumber in host byte order
 * @param hostname the name of the host to connect to
 * @param result the SOCKET (filled in)
 * @return OK if successful, SYSERR on failure
 **/
int initGNUnetClientSocket(unsigned short port,
			   char * hostname,
			   GNUNET_TCP_SOCKET * result) {
  struct hostent * he;

  he = gethostbyname(hostname); 
#if DEBUG_TCPIO
  LOG(LOG_DEBUG,
      "DEBUG: connecting to host '%s:%d'\n",
      hostname,
      port);
#endif
  if (he == NULL) {
    LOG(LOG_ERROR,
	"ERROR: could not find IP of host '%s'\n",
	hostname);
    return SYSERR;
  }  
  result->ip.addr.s_addr = ((struct in_addr*)he->h_addr)->s_addr;
  result->port = port;
  result->socket = -1; /* closed */
  result->outBufLen = 0;
  result->outBufPending = NULL;
  MUTEX_CREATE(&result->readlock);
  MUTEX_CREATE(&result->writelock);
  return OK;
}

/**
 * Initialize a GNUnet server socket.
 * @param sock the open socket
 * @param result the SOCKET (filled in)
 * @return OK (always successful)
 **/
int initGNUnetServerSocket(int sock,
			   GNUNET_TCP_SOCKET * result) {
  result->ip.addr.s_addr = 0;
  result->port = 0;
  result->socket = sock;
  result->outBufLen = 0;
  result->outBufPending = NULL;
  MUTEX_CREATE(&result->readlock);
  MUTEX_CREATE(&result->writelock);
  return OK;
}

/**
 * Check if a socket is open. Will ALWAYS return 'true'
 * for a valid client socket (even if the connection is
 * closed), but will return false for a closed server socket.
 * @return 1 if open, 0 if closed
 **/
int isOpenConnection(GNUNET_TCP_SOCKET * sock) {
  return (sock->socket != -1); 
}

/**
 * Check a socket, open and connect if it is closed and it is a client-socket.
 **/
int checkSocket(GNUNET_TCP_SOCKET * sock) {
  int res;
  struct sockaddr_in soaddr;

  if (sock->socket != -1)
    return OK;
  sock->socket = socket(PF_INET, SOCK_STREAM, 6); /* 6: TCP */
  if (sock->socket == -1) {
    LOG(LOG_FAILURE,
	"FAILURE: Can not create socket (%s).\n",
	strerror(errno));
    return SYSERR;
  }
  soaddr.sin_family = AF_INET;
  soaddr.sin_addr = sock->ip.addr;
  soaddr.sin_port = htons(sock->port);
  res = connect(sock->socket, 
		(struct sockaddr*)&soaddr,
		sizeof(soaddr));
  if ( (res < 0) &&
       (errno != EINPROGRESS) ) {
    LOG(LOG_INFO,
	"INFO: tcpio: Can not connect to %d.%d.%d.%d:%d (%s)\n",
	PRIP(ntohl(*(int*)&sock->ip.addr)),
	sock->port,
	strerror(errno));
    close(sock->socket);
    sock->socket = -1;
    return SYSERR;
  }
  return OK;
}

/**
 * Write to a GNUnet TCP socket.  Will also potentially complete the
 * sending of a previous non-blocking writeToSocket call.
 *
 * @param sock the socket to write to
 * @param buffer the buffer to write
 * @return OK if the write was sucessful, otherwise SYSERR.
 **/
int writeToSocket(GNUNET_TCP_SOCKET * sock,
		  CS_HEADER * buffer) {
  int res;
  int pos;
  int size;

  if (SYSERR == checkSocket(sock))
    return SYSERR;
  size = ntohs(buffer->size);
  pos = 0;

  MUTEX_LOCK(&sock->writelock);
  if (sock->outBufLen > 0) {
    while (sock->outBufLen > 0) {
#ifndef LINUX
      if (SIG_ERR == signal(SIGPIPE, SIG_IGN)) {
	if (SIG_ERR == signal(SIGPIPE, catcher))
	  LOG(LOG_WARNING,
	      "WARNING: could not install handler for SIGPIPE!\n"\
	      "Attempting to continue anyway.");
      }
      res = send(sock->socket,
		 sock->outBufPending,
		 sock->outBufLen,
		 0);
#else
      res = send(sock->socket,
		 sock->outBufPending,
		 sock->outBufLen,
		 MSG_NOSIGNAL);
#endif
      if (res < 0) {
	if (errno == EAGAIN) {
	  MUTEX_UNLOCK(&sock->writelock);
	  return SYSERR; /* would block, can not send right now;
			    but do NOT close socket in this case! */
	}
	LOG(LOG_INFO,
	    "INFO: TCP-write: send returned %d (%s)\n",
	    res, 
	    strerror(errno));
	closeSocketTemporarily(sock);
	MUTEX_UNLOCK(&sock->writelock);
	return SYSERR;
      }
      if (res < sock->outBufLen) {
	memcpy(sock->outBufPending,
	       &((char*)sock->outBufPending)[res],
	       sock->outBufLen - res);
	sock->outBufLen -= res;
      } else {
	FREE(sock->outBufPending);
	sock->outBufPending = NULL;
	sock->outBufLen = 0;
      }
    }      
  }

  while (pos < size) {
#ifndef LINUX
    if (SIG_ERR == signal(SIGPIPE, SIG_IGN)) {
      if (SIG_ERR == signal(SIGPIPE, catcher))
        LOG(LOG_WARNING,
          "WARNING: could not install handler for SIGPIPE!\n"\
          "Attempting to continue anyway.");
    }
    res = send(sock->socket,
	       &((char*)buffer)[pos],
	       size-pos,
	       0);
#else
    res = send(sock->socket,
	       &((char*)buffer)[pos],
	       size-pos,
	       MSG_NOSIGNAL);
#endif
    if (res < 0) {
      if (errno == EAGAIN) {
	MUTEX_UNLOCK(&sock->writelock);
	return SYSERR; /* would block, can not send right now;
			  but do NOT close socket in this case! */
      }
      LOG(LOG_INFO,
	  "INFO: TCP-write: send returned %d (%s)\n",
	  res, 
	  strerror(errno));
      closeSocketTemporarily(sock);
      MUTEX_UNLOCK(&sock->writelock);
      return SYSERR;
    }
    pos += res;
  }  
  MUTEX_UNLOCK(&sock->writelock);
  return OK;
}


/**
 * Write to a GNUnet TCP socket non-blocking.  Note that it is
 * possible that only a part of the message is send and that tcpio
 * buffers the rest until the next writeToSocket operation.  If that
 * buffer is full or if send did not transmit any byte of the message,
 * NO is returned to indicate that the write failed (would have
 * blocked).
 *
 * @param sock the socket to write to
 * @param buffer the buffer to write
 * @return OK if the write was sucessful, NO if it would have blocked and was not performed,
 *         otherwise SYSERR.
 **/
int writeToSocketNonBlocking(GNUNET_TCP_SOCKET * sock,
			     CS_HEADER * buffer) {
  int res;
  int size;

  if (SYSERR == checkSocket(sock))
    return SYSERR;
  MUTEX_LOCK(&sock->writelock);
  if (sock->outBufLen > 0) {
#ifndef LINUX
    if (SIG_ERR == signal(SIGPIPE, SIG_IGN)) {
      if (SIG_ERR == signal(SIGPIPE, catcher))
	LOG(LOG_WARNING,
	    "WARNING: could not install handler for SIGPIPE!\n"\
	    "Attempting to continue anyway.");
    }
#ifdef CYGWIN
    // enable non-blocking mode
    fcntl(sock, F_SETFL, fcntl(sock, F_GETFL) | O_NONBLOCK);
#endif
    res = send(sock->socket,
	       sock->outBufPending,
	       sock->outBufLen,
#ifndef CYGWIN
	       MSG_DONTWAIT);
#else
               0);
    fcntl(sock, F_SETFL, fcntl(sock, F_GETFL) & ~O_NONBLOCK);
#endif
#else
    res = send(sock->socket,
	       sock->outBufPending,
	       sock->outBufLen,
	       MSG_DONTWAIT|MSG_NOSIGNAL);
#endif
    if (res < 0) {
      if ( (errno == EWOULDBLOCK) ||
	   (errno == EAGAIN) ) {
	MUTEX_UNLOCK(&sock->writelock);
	return NO; 
      }
      LOG(LOG_INFO,
	  "INFO: TCP-write-nonblocking: send returned %d (%s)\n",
	  res, 
	  strerror(errno));
      closeSocketTemporarily(sock);
      MUTEX_UNLOCK(&sock->writelock);
      return SYSERR;
    }
    if (res < sock->outBufLen) {
      memcpy(sock->outBufPending,
	     &((char*)sock->outBufPending)[res],
	     sock->outBufLen - res);
      sock->outBufLen -= res;
      MUTEX_UNLOCK(&sock->writelock);
      return SYSERR;      
    }
    /* completely send out deferred buffer, so
       we can in fact continue! */
    FREENONNULL(sock->outBufPending);
    sock->outBufPending = NULL;
    sock->outBufLen = 0;
  }

  size = ntohs(buffer->size);

#ifndef LINUX
  if (SIG_ERR == signal(SIGPIPE, SIG_IGN)) {
    if (SIG_ERR == signal(SIGPIPE, catcher))
      LOG(LOG_WARNING,
          "WARNING: could not install handler for SIGPIPE!\n"\
          "Attempting to continue anyway.");
  }
#ifdef CYGWIN
    // enable non-blocking mode
    fcntl(sock, F_SETFL, fcntl(sock, F_GETFL) | O_NONBLOCK);
#endif
  res = send(sock->socket,
	     (char*)buffer,
	     size,
#ifndef CYGWIN
	     MSG_DONTWAIT);
#else
             0);
    fcntl(sock, F_SETFL, fcntl(sock, F_GETFL) & ~O_NONBLOCK);
#endif
#else
  res = send(sock->socket,
	     (char*)buffer,
	     size,
	     MSG_DONTWAIT|MSG_NOSIGNAL);
#endif
  if (res < 0) {
    if ( (errno == EWOULDBLOCK) ||
	 (errno == EAGAIN) ) {
      MUTEX_UNLOCK(&sock->writelock);
      return NO; /* would block, can not send right now;
		    but do NOT close socket in this case;
		    do not use SYSERR as return value
		    since this is not an error! */
    }
    LOG(LOG_INFO,
	"INFO: TCP-write-nonblocking: send returned %d (%s)\n",
	res, 
	strerror(errno));
    closeSocketTemporarily(sock);
    MUTEX_UNLOCK(&sock->writelock);
    return SYSERR;
  }
  if (res != size) {
    sock->outBufPending = MALLOC(size - res);
    memcpy(sock->outBufPending,
	   &((char*)buffer)[res],
	   size - res);
    sock->outBufLen = size - res;
    MUTEX_UNLOCK(&sock->writelock);
    return OK; /* return OK here means that the message will be transmitted,
		  though it may be a bit later (on the next call, in fact). */
  }
  MUTEX_UNLOCK(&sock->writelock);
  return OK;
}

/**
 * Read from a GNUnet TCP socket.
 * @param sock the socket
 * @param buffer the buffer to write data to
 * @return OK if the read was successful, SYSERR if the socket
 *         was closed by the other side (if the socket is a
 *         client socket and is used again, tcpio will attempt
 *         to re-establish the connection [temporary error]).
 **/
int readFromSocket(GNUNET_TCP_SOCKET * sock,
		   CS_HEADER ** buffer) {
  int res;
  int pos;
  char * buf;
  unsigned short size;

  if (SYSERR == checkSocket(sock))
    return SYSERR;
   
  MUTEX_LOCK(&sock->readlock);
  pos = 0;
  res = 0;
  while (pos < sizeof(unsigned short)) {

#ifdef LINUX
    res = recv(sock->socket,
	       &((char*)&size)[pos],
	       sizeof(unsigned short)-pos,
	       MSG_WAITALL|MSG_NOSIGNAL);
#else
 #ifdef CYGWIN
    res = recv(sock->socket,
	       &((char*)&size)[pos],
	       sizeof(unsigned short)-pos,
	       MSG_NOSIGNAL);
 #else
    res = recv(sock->socket,
	       &((char*)&size)[pos],
	       sizeof(unsigned short)-pos,
	       MSG_WAITALL);
 #endif
#endif

    if (res <= 0)
      break;    
    pos += res;
  }
  size = ntohs(size);
  if (pos != sizeof(unsigned short)) {
#if DEBUG_TCPIO    
    LOG(LOG_INFO,
	"INFO: TCP-read: read on socket failed (pos: %d, res: %d, socket: %d, sock %d, %s), closing\n",
	pos, 
	res, 
	sock->socket,
	(int) sock,
	strerror(errno));
#endif
    closeSocketTemporarily(sock);
    MUTEX_UNLOCK(&sock->readlock);
    return SYSERR; /* other side closed socket or invalid header */
  }
  buf = (char*) *buffer;
  if (buf == NULL) 
    buf = MALLOC(size);
  /*
  LOG(LOG_DEBUG,
      "DEBUG: reading %d bytes from socket\n",
      size);
  */
  while (pos < size) {
#ifdef LINUX
    res = recv(sock->socket,
	       &buf[pos],
	       size-pos,
	       MSG_WAITALL|MSG_NOSIGNAL);
#else
 #ifdef CYGWIN
    res = recv(sock->socket,
	       &buf[pos],
	       size-pos,
	       MSG_NOSIGNAL);
 #else
    res = recv(sock->socket,
	       &buf[pos],
	       size-pos,
	       MSG_WAITALL);
 #endif
#endif
    if (res < 0) {  /* error, abort */
      LOG(LOG_INFO,
	  "INFO: TCP-read: Invalid read on socket (%d, %s), closing\n",
	  res,
	  strerror(errno));
      closeSocketTemporarily(sock);
      if (*buffer == NULL)
	FREE(buf);
      MUTEX_UNLOCK(&sock->readlock);
      return SYSERR;
    }    
    pos += res;
    if (pos == size)
      break;
    if (res == 0) { /* other side closed connection, but we're not done yet */
      LOG(LOG_INFO,
	  "INFO: Other side closed connection (incomplete read (%d of %d), closing\n",
	  pos, 
	  size);
      closeSocketTemporarily(sock);
      if (*buffer == NULL)
	FREE(buf);
      MUTEX_UNLOCK(&sock->readlock);
      return SYSERR;
    }
  }
#if DEBUG_TCPIO
  LOG(LOG_DEBUG,
      "DEBUG: successfully received %d bytes from TCP socket\n",
      size);
#endif
  MUTEX_UNLOCK(&sock->readlock);
  *buffer = (CS_HEADER*) buf;
  (*buffer)->size = htons(size);
  return OK; /* success */
}

/**
 * Close a GNUnet TCP socket for now (use to temporarily close
 * a TCP connection that will probably not be used for a long
 * time; the socket will still be auto-reopened by the
 * readFromSocket/writeToSocket methods if it is a client-socket).
 **/
void closeSocketTemporarily(GNUNET_TCP_SOCKET * sock) {
  if (sock->socket != -1) {
    int i;

    i = sock->socket;
#if DEBUG_TCPIO
    LOG(LOG_DEBUG,
	"DEBUG: TCP: closing socket %d\n",
	sock->socket);
#endif
    sock->socket = -1;
    if (0 != shutdown(i, SHUT_RDWR)) 
      LOG(LOG_INFO,
	  "INFO: error shutting down socket %d: %s\n",
	  i,
	  strerror(errno));
    if (0 != close(i))
      LOG(LOG_INFO,
	  "INFO: error closing socket %d: %s\n",
	  i,
	  strerror(errno));
  }
  sock->outBufLen = 0;
  FREENONNULL(sock->outBufPending);
  sock->outBufPending = NULL;  
}

/**
 * Destroy a socket for good. If you use this socket afterwards,
 * you must first invoke initializeSocket, otherwise the operation
 * will fail.
 **/
void destroySocket(GNUNET_TCP_SOCKET * sock) {
  closeSocketTemporarily(sock);
  sock->ip.addr.s_addr = 0;
  sock->port = 0;
  sock->outBufLen = 0;
  FREENONNULL(sock->outBufPending);
  sock->outBufPending = NULL;
  MUTEX_DESTROY(&sock->readlock);
  MUTEX_DESTROY(&sock->writelock);
}


/*  end of tcpio.c */
